from typing import Iterable
from unittest import TestCase
from unittest.mock import Mock, call, sentinel

from google.protobuf.message import Message
from grpc import ServicerContext
from testfixtures import LogCapture

from frgal.exceptions import ServiceError
from frgal.service import wrap_stream, wrap_unary


class WrapUnaryTest(TestCase):
    def setUp(self):
        self.log_handler = LogCapture()

    def tearDown(self):
        self.log_handler.uninstall()

    def test_call_unary_unary(self):
        class TestServicer:
            @wrap_unary
            def serve(self, request: Message, context: ServicerContext) -> Message:
                return sentinel.response
        servicer = TestServicer()

        response = servicer.serve(sentinel.request, sentinel.context)

        self.assertEqual(response, sentinel.response)
        # Check logs
        self.assertEqual(len(self.log_handler.records), 2)
        self.assertRegex(self.log_handler.records[0].getMessage(), r'\[\w{4}\] .*TestServicer.serve(.*)')
        self.assertRegex(self.log_handler.records[1].getMessage(), r'\[\w{4}\] .*TestServicer.serve returned .*')

    def test_call_stream_unary(self):
        class TestServicer:
            @wrap_unary
            def serve(self, request: Iterable[Message], context: ServicerContext) -> Message:
                return sentinel.response
        servicer = TestServicer()

        response = servicer.serve([sentinel.request] * 5, sentinel.context)

        self.assertEqual(response, sentinel.response)
        # Check logs
        self.assertEqual(len(self.log_handler.records), 2)
        self.assertRegex(self.log_handler.records[0].getMessage(), r'\[\w{4}\] .*TestServicer.serve(.*)')
        self.assertRegex(self.log_handler.records[1].getMessage(), r'\[\w{4}\] .*TestServicer.serve returned .*')

    def test_service_error(self):
        class TestServicer:
            @wrap_unary
            def serve(self, request: Message, context: ServicerContext):
                raise ServiceError(sentinel.status, message=sentinel.message)
        servicer = TestServicer()
        context = Mock(ServicerContext)
        context.abort.side_effect = Exception('Abort!')

        with self.assertRaisesRegex(Exception, 'Abort!'):
            servicer.serve(sentinel.request, context)

        self.assertEqual(context.mock_calls, [call.abort(sentinel.status, sentinel.message)])
        # Check logs
        self.assertEqual(len(self.log_handler.records), 2)
        self.assertRegex(self.log_handler.records[0].getMessage(), r'\[\w{4}\] .*TestServicer.serve(.*)')
        self.assertRegex(self.log_handler.records[1].getMessage(), r'\[\w{4}\] .*TestServicer.serve failed with .*')

    def test_error(self):
        class TestServicer:
            @wrap_unary
            def serve(self, request: Message, context: ServicerContext):
                raise Exception('Smeghead!')
        servicer = TestServicer()

        with self.assertRaisesRegex(Exception, 'Smeghead!'):
            servicer.serve(sentinel.request, sentinel.context)

        # Check logs
        self.assertEqual(len(self.log_handler.records), 2)
        self.assertRegex(self.log_handler.records[0].getMessage(), r'\[\w{4}\] .*TestServicer.serve(.*)')
        self.assertRegex(self.log_handler.records[1].getMessage(), r'\[\w{4}\] .*TestServicer.serve raised error .*')


class WrapStreamTest(TestCase):
    def setUp(self):
        self.log_handler = LogCapture()

    def tearDown(self):
        self.log_handler.uninstall()

    def test_call_unary_stream(self):
        class TestServicer:
            @wrap_stream
            def serve(self, request: Message, context: ServicerContext) -> Iterable[Message]:
                for i in range(5):
                    yield sentinel.response
        servicer = TestServicer()

        response = servicer.serve(sentinel.request, sentinel.context)

        self.assertEqual(list(response), [sentinel.response] * 5)
        # Check logs
        self.assertEqual(len(self.log_handler.records), 6)
        self.assertRegex(self.log_handler.records[0].getMessage(), r'\[\w{4}\] .*TestServicer.serve(.*)')
        for i in range(5):
            self.assertRegex(self.log_handler.records[i + 1].getMessage(),
                             r'\[\w{4}\] .*TestServicer.serve returned .*')

    def test_call_stream_stream(self):
        class TestServicer:
            @wrap_stream
            def serve(self, request: Iterable[Message], context: ServicerContext) -> Iterable[Message]:
                for i in range(5):
                    yield sentinel.response
        servicer = TestServicer()

        response = servicer.serve([sentinel.request] * 5, sentinel.context)

        self.assertEqual(list(response), [sentinel.response] * 5)
        # Check logs
        self.assertEqual(len(self.log_handler.records), 6)
        self.assertRegex(self.log_handler.records[0].getMessage(), r'\[\w{4}\] .*TestServicer.serve(.*)')
        for i in range(5):
            self.assertRegex(self.log_handler.records[i + 1].getMessage(),
                             r'\[\w{4}\] .*TestServicer.serve returned .*')

    def test_service_error(self):
        class TestServicer:
            @wrap_stream
            def serve(self, request: Message, context: ServicerContext) -> Iterable[Message]:
                raise ServiceError(sentinel.status, message=sentinel.message)
        servicer = TestServicer()
        context = Mock(ServicerContext)
        context.abort.side_effect = Exception('Abort!')

        response = servicer.serve(sentinel.request, context)
        with self.assertRaisesRegex(Exception, 'Abort!'):
            list(response)

        self.assertEqual(context.mock_calls, [call.abort(sentinel.status, sentinel.message)])
        # Check logs
        self.assertEqual(len(self.log_handler.records), 2)
        self.assertRegex(self.log_handler.records[0].getMessage(), r'\[\w{4}\] .*TestServicer.serve(.*)')
        self.assertRegex(self.log_handler.records[1].getMessage(), r'\[\w{4}\] .*TestServicer.serve failed with .*')

    def test_error(self):
        class TestServicer:
            @wrap_stream
            def serve(self, request: Message, context: ServicerContext) -> Iterable[Message]:
                raise Exception('Smeghead!')
        servicer = TestServicer()

        response = servicer.serve(sentinel.request, sentinel.context)
        with self.assertRaisesRegex(Exception, 'Smeghead!'):
            list(response)

        # Check logs
        self.assertEqual(len(self.log_handler.records), 2)
        self.assertRegex(self.log_handler.records[0].getMessage(), r'\[\w{4}\] .*TestServicer.serve(.*)')
        self.assertRegex(self.log_handler.records[1].getMessage(), r'\[\w{4}\] .*TestServicer.serve raised error .*')
